import logging

import pandas as pd
pd.options.mode.chained_assignment = None  # default='warn'

from typing import Union

from matplotlib import pyplot as plt
from sklearn.base import BaseEstimator
from sklearn.metrics import classification_report, roc_curve
from sklearn.pipeline import Pipeline

from .builders import buildMetaPipeline
from .Models import BestModel
from .Models import DemocraticModel
from .Models import OneRulerForAll


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class MetaPipeline(BaseEstimator):
    #Wrapper

    def __init__(self, model: Union[BestModel,DemocraticModel,OneRulerForAll], verbose: bool=True):
        self.model = model
        self.pipe : Pipeline
        self.verbose = verbose
        # To shut the logs
        if not verbose: logging.basicConfig(level=logging.CRITICAL)

    def fit(self, X: pd.DataFrame, y: pd.Series) -> BaseEstimator:

        # some of the MetaPipeline steps requires information on the data, therefore we have to initialize it here
        self.pipe = buildMetaPipeline(X, self.model)

        self.pipe.fit(X, y)

        return self

    #Overloading BaseEstimator methods
    def predict(self, X: pd.DataFrame):
        return self.pipe.predict(X)

    def predict_proba(self, X: pd.DataFrame):
        return self.pipe.predict_proba(X)

    def transform(self, X: pd.DataFrame, y=None):
        return self.pipe.transform(X)

    def fit_transform(self, X: pd.DataFrame, y: pd.Series):
        return self.pipe.fit(X, y).transform(X)

    def get_scores(self, X : pd.DataFrame, y: pd.Series):
            return self.pipe.named_steps[self.model.__repr__()].estimatorPool.get_scores(self.pipe.transform(X),y)

    def classification_report(self, X: pd.DataFrame, y: pd.Series):
        #Return sklearn classification report
        y_pred = self.pipe.predict(X)
        print(classification_report(y, y_pred))

    def roc_curve(self, X: pd.DataFrame, y: pd.Series):

        ns_probs = [0 for _ in range(len(y))]

        # predict probabilities
        lr_probs = self.pipe.predict_proba(X)
        # keep probabilities for the positive outcome only
        lr_probs = lr_probs[:, 1]
        # calculate roc curves
        ns_fpr, ns_tpr, _ = roc_curve(y, ns_probs)
        lr_fpr, lr_tpr, _ = roc_curve(y, lr_probs)
        # plot the roc curve for the model
        plt.plot(ns_fpr, ns_tpr, linestyle='--', label='No Skill')
        plt.plot(lr_fpr, lr_tpr, marker='.', label='Logistic')
        # axis labels
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        # show the legend
        plt.legend()
        # show the plot
        plt.show()





import numpy as np

class PenalizationGrid:

    def __init__(self, minCoef=1e-10, maxCoef=1, length=200):
        self.values = np.linspace(maxCoef, minCoef, length).tolist()

    def isEmpty(self)->bool:
        return len(self.values)==0

    def getNextKCoeffs(self,k):
        penalizationCoeffsForBatch = self.values[:k]
        del self.values[:k]
        return penalizationCoeffsForBatch

    def size(self):
        return len(self.values)



import logging

import pandas as pd

from sklearn.base import BaseEstimator


from .EstimatorsPool import EstimatorPool
from ..support.MyTools import getAdaptedCrossVal, checkClassBalance


class BestModel(BaseEstimator):

    def __init__(self, gridSearch: bool=True, metrics: str='accuracy', n_splits: int =10):
        self.best_estimator_name : str
        self.best_estimator : BaseEstimator
        self.best_estimator_index : int

        # Pool of estimators
        self.estimatorPool = EstimatorPool()
        self.scores = pd.DataFrame()
        self.n_splits = n_splits
        self.grid_search = gridSearch
        self.metrics = metrics

    def fit(self, X: pd.DataFrame, y: pd.Series, ) -> BaseEstimator:


        logging.info("Training models")

        # Pour détecter une distribution déséquilibrée...
        checkClassBalance(y)
        # Récupération d'un split de CV adapté selon l'indexage du set
        cv = getAdaptedCrossVal(X, self.n_splits)

        if self.grid_search:
            self.estimatorPool.fitWithGridSearch(X,y,cv,'accuracy')
        else:
            self.estimatorPool.fit(X, y)

        # Getting the best estimator according to the metric mean
        best_score , self.best_estimator_name, self.best_estimator = self.estimatorPool.get_best(X,y)

        logging.info("The best estimator is {0} with a cross-validation accuracy (in Sample) of {1}".format(
            self.best_estimator_name, best_score))

        return self

    # Overloading sklearn BaseEstimator methods to use the best estimator
    def predict(self, X: pd.Series) -> pd.Series:
        return self.best_estimator.predict(X)

    def predict_proba(self, X: pd.Series) -> pd.Series:
        return self.best_estimator.predict_proba(X)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X

    def __repr__(self, **kwargs):
        return 'Meta Model'



from lib2to3.pytree import Base
from typing import Union, Tuple, Any

import numpy as np
import pandas as pd
from numpy import ndarray
from sklearn.base import BaseEstimator
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score
from sklearn.model_selection import (RandomizedSearchCV, StratifiedKFold,
                                     TimeSeriesSplit)
from sklearn.naive_bayes import GaussianNB
from ..constants.gsp import estimators_params


class EstimatorPool(BaseEstimator):

    def __init__(self):

        self.estimatorsList = [("random forest classifier", RandomForestClassifier()),
                               ("Logistic Regression", LogisticRegression(fit_intercept=True)),
                               ('Gaussian Naive Bayes', GaussianNB()),
                               ('LDA', LinearDiscriminantAnalysis()),
                               ]

    def fit(self, X: pd.DataFrame, y: pd.Series) -> list[tuple[str, BaseEstimator]]:
        for estimator in self.estimatorsList: estimator[1].fit(X, y)
        return self.estimatorsList

    def fitWithGridSearch(self, X: pd.DataFrame, y: pd.Series,
                          cv: Union[TimeSeriesSplit, StratifiedKFold],
                          metrics) -> list[tuple[str, BaseEstimator]]:

        for estimator in self.estimatorsList:
            if estimator[0] in estimators_params:
                clf = RandomizedSearchCV(estimator=estimator[1],
                                         param_distributions=estimators_params[estimator[0]], scoring=metrics,
                                         n_jobs=-1, cv=cv)
                clf.fit(X, y)

                estimator[1].set_params(**clf.best_params_)

            estimator[1].fit(X, y)

        return self.estimatorsList

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {estimator[0]: estimator[1].predict(X) for estimator in self.estimatorsList})
        
    def predict_proba(self, X: pd.DataFrame) -> ndarray:
        return np.array([estimator[1].predict_proba(X) for estimator in self.estimatorsList])

    def get_best(self, X: pd.DataFrame, y: pd.Series) -> tuple[float, str, BaseEstimator]:

        scores = [accuracy_score(estimator[1].predict(X), y) for estimator in self.estimatorsList]
        return float(np.max(scores)), *self.estimatorsList[np.argmax(scores)]

    def get_scores(self, X: pd.DataFrame, y: pd.Series) -> list[tuple[str, float]]: 
        return [(estimator_name, accuracy_score(estimator.predict(X), y)) for estimator_name,estimator in self.estimatorsList]
        
    def __len__(self):
        return len(self.estimatorsList)



import logging
import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.feature_selection import SelectKBest

from .LassoFeatureSelection import FeatureSelection


class LassoSelectorTransformer(BaseEstimator, TransformerMixin):
    """
    Selects features according to their order of apparition when varying the shrinkage coefficient of LASSO regression
    """

    def __init__(self, preSelectionSize=50):

        self.selectedFeaturesNames = list[str]
        self.preSelectionSize = preSelectionSize

    def __preselect(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        # Preselection for datasets with many features
        preselector = SelectKBest(k=self.preSelectionSize)
        preselector.fit(X, y)
        cols = preselector.get_support(indices=True).tolist()
        X = X.copy()
        return X.iloc[:, cols]

    def fit(self, X: pd.DataFrame, y: pd.Series) -> TransformerMixin:
        X = self.__preselect(X, y) if (X.shape[1]) > self.preSelectionSize else X

        featureSelection = FeatureSelection()
        featureSelection.fit(X, y)

        self.selectedFeaturesNames = featureSelection.getSelectedFeaturesNames()
        return self

    def transform(self, X: pd.DataFrame, y=None):
        X = X.copy()
        
        return X.loc[:,self.selectedFeaturesNames]



import logging
from typing import Optional

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier

from  .EstimatorsPoolCV import EstimatorPoolCV

from  ..support.MyTools import getAdaptedCrossVal, checkClassBalance

class OneRulerForAllCV(BaseEstimator):
    """
    The OneRulerForAll bottleneck estimator uses stacking techniques to leverage the strengths of estimators in the pool.
    We first train a pool of estimator, then another 'ruler' estimator is trained to decide which estimator in
    the pool might be right, given the pool outputs
    """

    def __init__(self, gridSearch: bool = True, metrics: str = 'accuracy', nSplits: int=10, ruler: Optional[BaseEstimator] =None):

        if ruler is None:
            self.ruler = RandomForestClassifier()
            self.rulerName = 'random forest classifier'
        else:
            self.ruler = ruler

        self.estimatorPoolCV = EstimatorPoolCV()
        self.nSplits = nSplits
        self.gridSearch = gridSearch
        self.metrics = metrics

    def fit(self, X: pd.DataFrame, y: pd.Series) -> BaseEstimator:

        checkClassBalance(y)

        logging.info("Training models...")

        cv = getAdaptedCrossVal(X, self.nSplits)

        # Training the pool
        if self.gridSearch:
            self.estimatorPoolCV.fitWithGridSearch(X, y, cv, self.metrics)
        else:
            self.estimatorPoolCV.fit(X, y)

        estimatorsPoolOutputs = self.estimatorPoolCV.predict(X)

        self.ruler.fit(estimatorsPoolOutputs, y)

        return self

    # Overriding sklearn BaseEstimator methods
    def predict(self, X: pd.DataFrame) -> pd.Series:
        estimatorsPoolOutputs = self.estimatorPoolCV.predict(X)
        return self.ruler.predict(estimatorsPoolOutputs)

    def predict_proba(self, X: pd.DataFrame) -> pd.Series:
        estimatorsPoolOutputs = self.estimatorPoolCV.predict(X)
        return self.ruler.predict_proba(estimatorsPoolOutputs)

    def transform(self, X: pd.DataFrame):
        return X

    def __repr__(self, **kwargs):
        return 'ORFA CV'



import logging

import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator

from .EstimatorsPoolCV import EstimatorPoolCV

from ..support.MyTools import getAdaptedCrossVal, checkClassBalance

class _AvailableIfDescriptor:
    """Implements a conditional property using the descriptor protocol.
    Using this class to create a decorator will raise an ``AttributeError``
    if check(self) returns a falsey value. Note that if check raises an error
    this will also result in hasattr returning false.
    See https://docs.python.org/3/howto/descriptor.html for an explanation of
    descriptors.
    """

    def __init__(self, fn, check, attribute_name):
        self.fn = fn
        self.check = check
        self.attribute_name = attribute_name

        # update the docstring of the descriptor

    def __get__(self, obj, owner=None):
        attr_err = AttributeError(
            f"This {repr(owner.__name__)} has no attribute {repr(self.attribute_name)}"
        )
        if obj is not None:
            # delegate only on instances, not the classes.
            # this is to allow access to the docstrings.
            if not self.check(obj):
                raise attr_err

            # lambda, but not partial, allows help() to work with update_wrapper
            out = lambda *args, **kwargs: self.fn(obj, *args, **kwargs)  # noqa
        else:

            def fn(*args, **kwargs):
                if not self.check(args[0]):
                    raise attr_err
                return self.fn(*args, **kwargs)

            # This makes it possible to use the decorated method as an unbound method,
            # for instance when monkeypatching.
            out = lambda *args, **kwargs: fn(*args, **kwargs)  # noqa
        return out


def available_if(check):
    return lambda fn: _AvailableIfDescriptor(fn, check, attribute_name=fn.__name__)


class DemocraticModelCV(BaseEstimator):
    """
    Hard Voting Classifier
    
    The Democratic bottleneck estimator makes the trained models vote to decide the output. 
    Classes are assumed to be in alphabetical order
    """

    def __init__(self, gridSearch: bool = True, metrics: str = 'accuracy', nSplits: int =10, voting: str='soft'):
        self.estimatorPoolCV = EstimatorPoolCV()
        self.nSplits = nSplits
        self.gridSearch = gridSearch
        self.metrics = metrics
        self.voting = voting
        self.nEstimators = len(self.estimatorPoolCV)

    def _check_soft_voting(self):
        if self.voting == "hard":
            raise AttributeError(
                f"predict_proba is not available when voting={self.voting}"
            )
        return True

    def _check_hard_voting(self):
        if self.voting == "soft":
            raise AttributeError(
                f"predict_proba is not available when voting={self.voting}"
            )
        return True

    def fit(self, X: pd.DataFrame, y: pd.Series) -> BaseEstimator:

        checkClassBalance(y)
        logging.info("Training models...")
        cv = getAdaptedCrossVal(X, self.nSplits)

        # Training the pool
        if self.gridSearch:
            self.estimatorPoolCV.fitWithGridSearch(X, y, cv, self.metrics)
        else:
            self.estimatorPoolCV.fit(X, y)

        return self

    # Overriding sklearn BaseEstimator methods
    def predict(self, X: pd.DataFrame) -> np.ndarray:

        return np.argmax(self.predict_proportion(X), axis=1) if self.voting == 'hard' else np.argmax(
            self.predict_proba(X),axis=1)

    @available_if(_check_hard_voting)
    def predict_proportion(self, X: pd.DataFrame) -> np.ndarray:
        """
        Returns the proportion of model votes for each class
        """
        estimatorsPoolOutputs = self.estimatorPoolCV.predict(X)
        estimator_names = estimatorsPoolOutputs.columns
        classes = np.sort(np.unique(estimatorsPoolOutputs.values))

        for c in classes:
            estimatorsPoolOutputs[f"{c}_proportion"] = (estimatorsPoolOutputs[estimator_names] == c).sum(
                axis=1) / self.nEstimators

        return estimatorsPoolOutputs[[f"{c}_proportion" for c in classes]].values

    @available_if(_check_soft_voting)
    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        """
        Returns the average model probability per class
        """
        estimatorsPoolProbas = self.estimatorPoolCV.predict_proba(X)
        return np.mean(estimatorsPoolProbas, axis=0)

    def transform(self, X: pd.DataFrame):
        return X

    def __repr__(self, **kwargs):
        return 'Democratic Model CV'



import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler, OneHotEncoder, FunctionTransformer

from .Preprocessing.LassoSelectorTransformer import LassoSelectorTransformer
from .Preprocessing.NonStationarityCorrector import NonStationarityCorrector
from .support.MyTools import isIndexedByTime


def buildColumnTransformer(X: pd.DataFrame) -> ColumnTransformer:

    # Select numerical and categorical feature to be able to apply different transformers

    numerical_ix = X.select_dtypes(include=['int64', 'float64']).columns
    categorical_ix = X.select_dtypes(include=['object', 'bool']).columns

    if isIndexedByTime(X):
        numerical_process = Pipeline([('NonStationarityCorrector', NonStationarityCorrector()),
                                      ('MinMaxScaler', MinMaxScaler(feature_range=[-1, 1]))])
    else:
        numerical_process = Pipeline([('StandardScaler', StandardScaler()),
                                      ('MinMaxScaler', MinMaxScaler(feature_range=[-1, 1]))])


    transformer = [('Categorical', OneHotEncoder(), categorical_ix),
                   ("Numerical", numerical_process, numerical_ix)]

    return ColumnTransformer(transformers=transformer)


def buildMetaPipeline(X: pd.DataFrame, estimator: BaseEstimator) -> Pipeline:
    cols = X.columns
    columnTransformer = buildColumnTransformer(X)
    
    if len(cols) > 15:
        return Pipeline(
            [('Preprocessing', columnTransformer),
             ('fc_tf', FunctionTransformer(lambda x: pd.DataFrame(x, columns=cols))),
             ("Lasso Selector", LassoSelectorTransformer()),
             (estimator.__repr__(), estimator)])
    else:
        return Pipeline(
            [('Preprocessing', columnTransformer),
             ('fc_tf', FunctionTransformer(lambda x: pd.DataFrame(x, columns=cols))),
             (estimator.__repr__(), estimator)])


from typing import Union

import numpy as np
import pandas as pd
from numpy import ndarray
from sklearn.base import BaseEstimator
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score
from sklearn.model_selection import (RandomizedSearchCV, StratifiedKFold,
                                     TimeSeriesSplit)
from sklearn.naive_bayes import GaussianNB
from sklearn.pipeline import Pipeline
from tqdm import tqdm
from  ..constants.gsp import estimators_params
from  ..builders import buildMetaPipeline


class EstimatorPoolCV(BaseEstimator):

    def __init__(self):

        self.estimatorsList = [("random forest classifier", RandomForestClassifier()),
                               ("Logistic Regression", LogisticRegression(fit_intercept=True)),
                               ('Gaussian Naive Bayes', GaussianNB()),
                               ('LDA', LinearDiscriminantAnalysis()),
                               ]

        self.estimatorsPipeline: list[tuple[str,Pipeline]] = []

    def fit(self, X: pd.DataFrame, y: pd.Series) -> list[tuple[str, Pipeline]]:
        self.estimatorsPipeline = [(estimator_name, buildMetaPipeline(X, estimator).fit(X,y)) for estimator_name, estimator in self.estimatorsList]

        return self.estimatorsPipeline

    def fitWithGridSearch(self, X: pd.DataFrame, y: pd.Series,
                          cv: Union[TimeSeriesSplit, StratifiedKFold],
                          metrics) -> list[tuple[str, Pipeline]]:

        for estimator in tqdm(self.estimatorsList):
            pipe = buildMetaPipeline(X,estimator=estimator[1])
            if estimator[0] in estimators_params:
                grid = {f'{estimator[1].__repr__()}__{key}': value for key, value in estimators_params[estimator[0]].items()}
                clf = RandomizedSearchCV(estimator=pipe,
                                         param_distributions=grid, scoring=metrics,
                                         n_jobs=-1, cv=cv, verbose=4)
                clf.fit(X, y)

                pipe.set_params(**clf.best_params_)

            pipe.fit(X, y)
            self.estimatorsPipeline.append((estimator[0], pipe))

        return self.estimatorsPipeline

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            {estimator_name[0]: pipe.predict(X) for estimator_name, pipe in self.estimatorsPipeline})
        
    def predict_proba(self, X: pd.DataFrame) -> ndarray:
        return np.array([pipe.predict_proba(X) for _,pipe in self.estimatorsPipeline])

    def get_best(self, X: pd.DataFrame, y: pd.Series) -> tuple[float, str, Pipeline]:

        scores = [accuracy_score(pipe.predict(X), y) for _,pipe in self.estimatorsPipeline]
        return float(np.max(scores)), *self.estimatorsPipeline[np.argmax(scores)]

    def get_scores(self, X: pd.DataFrame, y: pd.Series) -> list[tuple[str, float]]: 
        return [(estimator_name, accuracy_score(pipe.predict(X), y)) for estimator_name,pipe in self.estimatorsPipeline]


    def __len__(self):
        return len(self.estimatorsPipeline)



import threading

import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from .PenalizationGrid import PenalizationGrid


class FeatureSelection:

    def __init__(self, batchSize=10, nbFeatureToSelect=15):

        self.penalizationGrid = PenalizationGrid()
        self.regressorsCoeffsValues = []
        assert self.penalizationGrid.size() % batchSize == 0, 'The batch size must divide the size of the ' \
                                                              'penalization coefficients grid. '
        self.batchSize = batchSize
        self.nbFeatureToSelect = nbFeatureToSelect
        self.selectedFeaturesNames = None

    def __featureSelectionStep(self, penalizationCoeff: float, X, y) -> None:
        lasso = LogisticRegression(C=penalizationCoeff, fit_intercept=True, penalty='l1', solver='saga', max_iter=4000,
                                   verbose=0)
        lasso.fit(X, y)
        self.regressorsCoeffsValues.append(lasso.coef_[0])

    def __featureSelectionBatchStep(self, penalizationPartialGrid: list, X, y):
        threads = [None] * self.batchSize

        for threadIndex, penalizationCoeff in enumerate(penalizationPartialGrid):
            threads[threadIndex] = threading.Thread(target=self.__featureSelectionStep, args=[penalizationCoeff, X, y])
            threads[threadIndex].start()

        for thread in threads:
            thread.join()

    def fit(self, X: pd.DataFrame, y: pd.Series):

        while not self.penalizationGrid.isEmpty():
            penalizationPartialGrid = self.penalizationGrid.getNextKCoeffs(self.batchSize)
            self.__featureSelectionBatchStep(penalizationPartialGrid, X, y)

        indexForObjNbOfFeatures = min(range(len(self.regressorsCoeffsValues)), key=lambda index: np.abs(
            sum(self.regressorsCoeffsValues[index] != 0) - self.nbFeatureToSelect))

        self.selectedFeaturesNames = X.columns[self.regressorsCoeffsValues[indexForObjNbOfFeatures] != 0].values.tolist()

    def getSelectedFeaturesNames(self) -> list[str]:
        return self.selectedFeaturesNames



import logging

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler
from statsmodels.tsa.stattools import adfuller
from ..constants.GLOBAL_PARAMS import WINDOW


class NonStationarityCorrector(BaseEstimator, TransformerMixin):
    """
    Correct columns that don't seem stationary according to Augmented Dicker Fuller statistical stationarity test
    """

    def __init__(self, colsToKeepIntact=None):
        # If the user wants some columns to remain intact, they can be passed as arguments
        if colsToKeepIntact is None:
            colsToKeepIntact = []
        self.colsToKeepIntact = colsToKeepIntact
        self.colsToCorrect = []

    def fit(self, X: pd.DataFrame, y: pd.Series) -> TransformerMixin:
        '''
        split the cols according to Augmented Dicker Fuller statistical stationarity test
        '''
        for col in X.columns:
            if col not in self.colsToKeepIntact:
                test_results = adfuller(X[col].values)[0]
                if test_results < 0:
                    self.colsToCorrect.append(col)
                else:
                    self.colsToKeepIntact.append(col)

        assert len(self.colsToKeepIntact) + len(self.colsToCorrect) == len(X.columns), "Column(s) were lost in process"
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        # Apply the transformations
        X = X.copy()
        logging.debug("Correcting non-stationarity on the dataset...")
        if X.shape[0] > WINDOW:
            for col in self.colsToCorrect:
                # Depending on the window size, std can be null.
                # In that situation the actual value can be replaced by the last non null value
                # We also use loc[start+WINDOW:] in order to leave the WINDOW first rows intact. Otherwise, it would be nans
                X[col].iloc[WINDOW:] = ((X[col] - X[col].rolling(window=WINDOW).mean()) / X[col].rolling(window=WINDOW).std(skipna=True).replace(to_replace=0., method='ffill')).iloc[WINDOW:]

            if self.colsToKeepIntact: X[self.colsToKeepIntact] = StandardScaler().fit_transform(
                X[self.colsToKeepIntact])
        else:
            X[X.columns] = StandardScaler().fit_transform(X[X.columns])
            assert type(X) == pd.DataFrame, 'type error'

        return X



import logging
from datetime import datetime
from typing import Union

import numpy as np
import pandas as pd
from sklearn.model_selection import TimeSeriesSplit, StratifiedKFold


def extract_score_params(cv_results: dict, n_splits: int, k_last_splits=5) -> tuple[np.float64, dict]:
    """
    :param k_last_splits: k last splits to consider for averaging scores
    :param cv_results: output of GridSearchCV.cv_results_
    :param n_splits: number of splits in the cross_validation
    :return: best score on the last k splits, and the params
    """
    if n_splits < k_last_splits:
        logging.warning(
            'There is more k_last_splits to include than the number of total splits.'
            ' k_last_splits will be set to {0}'.format(
                n_splits))

    idx_best_config = np.argmin(cv_results['rank_test_score'])
    score = np.mean([cv_results['split{0}_test_score'.format(k)][idx_best_config] for k in
                     range(n_splits, n_splits - k_last_splits)])[0]
    params = cv_results['params'][idx_best_config]
    return score, params


def isIndexedByTime(X: pd.DataFrame) -> bool:
    return X.index.dtype in [datetime, 'datetime64[ns]']


def getAdaptedCrossVal(X: pd.DataFrame, nSplits: int) -> Union[TimeSeriesSplit, StratifiedKFold]:
    if isIndexedByTime(X):
        # Pour des données temporelles, on préféra découper linéairement dans le temps
        return TimeSeriesSplit(n_splits=nSplits)

    else:
        # Si les données ne sont pas temporelles, on applique un découpage stratifié en prenant des
        return StratifiedKFold(n_splits=nSplits)


def checkClassBalance(y: pd.Series):
    if len(y[y == 1]) / len(y) > 0.7:
        logging.warning('Unbalanced Target')
        raise(ValueError)
    return True



import logging
from typing import Optional

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import RandomizedSearchCV

from .EstimatorsPool import EstimatorPool
from  ..support.MyTools import getAdaptedCrossVal, checkClassBalance
from  ..constants.gsp import estimators_params


class OneRulerForAll(BaseEstimator):
    """
    The OneRulerForAll bottleneck estimator uses stacking techniques to leverage the strengths of estimators in the pool.
    We first train a pool of estimator, then another 'ruler' estimator is trained to decide which estimator in
    the pool might be right, given the pool outputs
    """

    def __init__(self, gridSearch: bool = True, metrics: str = 'accuracy', nSplits: int=10, ruler: Optional[BaseEstimator] =None):

        if ruler is None:
            self.ruler = RandomForestClassifier()
            self.rulerName = 'random forest classifier'
        else:
            self.ruler = ruler

        self.estimatorPool = EstimatorPool()
        self.nSplits = nSplits
        self.gridSearch = gridSearch
        self.metrics = metrics

    def fit(self, X: pd.DataFrame, y: pd.Series) -> BaseEstimator:

        checkClassBalance(y)

        logging.info("Training models...")

        cv = getAdaptedCrossVal(X, self.nSplits)

        # Training the pool
        if self.gridSearch:
            self.estimatorPool.fitWithGridSearch(X, y, cv, self.metrics)
        else:
            self.estimatorPool.fit(X, y)

        estimatorsPoolOutputs = self.estimatorPool.predict(X)

        # Training the ruler, with gridSearch if possible
        if self.rulerName in estimators_params:
            clf = RandomizedSearchCV(estimator=RandomForestClassifier(),
                                     param_distributions=estimators_params[self.rulerName], scoring=self.metrics,
                                     n_jobs=-1, cv=cv)
            clf.fit(estimatorsPoolOutputs, y)
            self.ruler.set_params(**clf.best_params_)
        self.ruler.fit(estimatorsPoolOutputs, y)

        return self

    # Overriding sklearn BaseEstimator methods
    def predict(self, X: pd.DataFrame) -> pd.Series:
        estimatorsPoolOutputs = self.estimatorPool.predict(X)
        return self.ruler.predict(estimatorsPoolOutputs)

    def predict_proba(self, X: pd.DataFrame) -> pd.Series:
        estimatorsPoolOutputs = self.estimatorPool.predict(X)
        return self.ruler.predict_proba(estimatorsPoolOutputs)

    def transform(self, X: pd.Series):
        return X

    def __repr__(self, **kwargs):
        return 'ORFA'



from setuptools import setup
 
setup(
    name='TinyAutoML',
    version='0.2.2.2',
    packages=['TinyAutoML', 'TinyAutoML.Preprocessing', 'TinyAutoML.constants', 'TinyAutoML.Models', 'TinyAutoML.support'],
    url='https://github.com/g0bel1n/TinyAutoML',
    license='MIT',
    author='g0bel1n',
    author_email='lucas.saban@ensae.fr',
    install_requires = ['pandas',
                        'scikit-learn',
                        'tqdm',
                        'numpy',
                        'statsmodels',
                        'matplotlib',
                        'xgboost'],
    description='Combinaison of ML models for binary classification. Academic Project.',
    download_url = 'https://github.com/g0bel1n/TinyAutoML/archive/refs/tags/v0.2.2.2.tar.gz',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',

    ],
)



from .DemocraticModel import DemocraticModel
from .BestModel import BestModel
from .OneRulerForAll import OneRulerForAll

from .DemocraticModelCV import DemocraticModelCV
from .BestModelCV import BestModelCV
from .OneRulerForAllCV import OneRulerForAllCV

__all__ = ['DemocraticModel','BestModel','OneRulerForAll', 'DemocraticModelCV', 'BestModelCV', 'OneRulerForAllCV']


# RANDOM FOREST

n_estimators = [3, 10, 30]
# Number of features to consider at every split
max_features = ['auto', 'sqrt']
# Maximum number of levels in tree
max_depth = [2, 3, 4, 5]

# Method of selecting samples for training each tree
bootstrap = [False]
# Create the random grid
random_forest_grid = {'n_estimators': n_estimators,
                      'max_features': max_features,
                      'max_depth': max_depth,
                      'bootstrap': bootstrap}

xgb_params = {'objective':['binary:logistic'],
        'min_child_weight': [1, 5, 10],
        'gamma': [0.5, 1, 1.5, 2, 5],
        'subsample': [0.6, 0.8, 1.0],
        'colsample_bytree': [0.6, 0.8, 1.0],
        'max_depth': [3, 4, 5],
        'silent' : [True],
        'verbosity' : [0]
        }

estimators_params = {'random forest classifier': random_forest_grid, "xgb": xgb_params}
